@@ -1,8 +1,11 @@
+#include "ares_setup.h"
+#include "ares.h"
+#include "ares_nameser.h"
 #include "ares-test.h"
+#include "ares-test-ai.h"
 #include "dns-proto.h"
 
 // Include ares internal files for DNS protocol details
-#include "nameser.h"
 #include "ares_dns.h"
 
 #ifdef HAVE_NETDB_H
@@ -19,11 +22,9 @@
 
 #ifdef WIN32
 #define BYTE_CAST (char *)
-#define sclose(x) closesocket(x)
 #define mkdir_(d, p) mkdir(d)
 #else
 #define BYTE_CAST
-#define sclose(x) close(x)
 #define mkdir_(d, p) mkdir(d, p)
 #endif
 
@@ -31,7 +32,31 @@ namespace ares {
 namespace test {
 
 bool verbose = false;
-int mock_port = 5300;
+static constexpr int dynamic_port = 0;
+int mock_port = dynamic_port;
+
+const std::vector<int> both_families = {AF_INET, AF_INET6};
+const std::vector<int> ipv4_family = {AF_INET};
+const std::vector<int> ipv6_family = {AF_INET6};
+
+const std::vector<std::pair<int, bool>> both_families_both_modes = {
+  std::make_pair<int, bool>(AF_INET, false),
+  std::make_pair<int, bool>(AF_INET, true),
+  std::make_pair<int, bool>(AF_INET6, false),
+  std::make_pair<int, bool>(AF_INET6, true)
+};
+const std::vector<std::pair<int, bool>> ipv4_family_both_modes = {
+  std::make_pair<int, bool>(AF_INET, false),
+  std::make_pair<int, bool>(AF_INET, true)
+};
+const std::vector<std::pair<int, bool>> ipv6_family_both_modes = {
+  std::make_pair<int, bool>(AF_INET6, false),
+  std::make_pair<int, bool>(AF_INET6, true)
+};
+
+// Which parameters to use in tests
+std::vector<int> families = both_families;
+std::vector<std::pair<int, bool>> families_modes = both_families_both_modes;
 
 unsigned long long LibraryTest::fails_ = 0;
 std::map<size_t, int> LibraryTest::size_fails_;
@@ -112,7 +137,7 @@ bool LibraryTest::ShouldAllocFail(size_t size) {
 
 // static
 void* LibraryTest::amalloc(size_t size) {
-  if (ShouldAllocFail(size)) {
+  if (ShouldAllocFail(size) || size == 0) {
     if (verbose) std::cerr << "Failing malloc(" << size << ") request" << std::endl;
     return nullptr;
   } else {
@@ -147,8 +172,8 @@ void DefaultChannelModeTest::Process() {
   ProcessWork(channel_, NoExtraFDs, nullptr);
 }
 
-MockServer::MockServer(int family, int port, int tcpport)
-  : udpport_(port), tcpport_(tcpport ? tcpport : udpport_), qid_(-1) {
+MockServer::MockServer(int family, int port)
+  : udpport_(port), tcpport_(port), qid_(-1) {
   // Create a TCP socket to receive data on.
   tcpfd_ = socket(family, SOCK_STREAM, 0);
   EXPECT_NE(-1, tcpfd_);
@@ -175,18 +200,48 @@ MockServer::MockServer(int family, int port, int tcpport)
     addr.sin_port = htons(udpport_);
     int udprc = bind(udpfd_, (struct sockaddr*)&addr, sizeof(addr));
     EXPECT_EQ(0, udprc) << "Failed to bind AF_INET to UDP port " << udpport_;
+    // retrieve system-assigned port
+    if (udpport_ == dynamic_port) {
+      ares_socklen_t len = sizeof(addr);
+      auto result = getsockname(udpfd_, (struct sockaddr*)&addr, &len);
+      EXPECT_EQ(0, result);
+      udpport_ = ntohs(addr.sin_port);
+      EXPECT_NE(dynamic_port, udpport_);
+    }
+    if (tcpport_ == dynamic_port) {
+      ares_socklen_t len = sizeof(addr);
+      auto result = getsockname(tcpfd_, (struct sockaddr*)&addr, &len);
+      EXPECT_EQ(0, result);
+      tcpport_ = ntohs(addr.sin_port);
+      EXPECT_NE(dynamic_port, tcpport_);
+    }
   } else {
     EXPECT_EQ(AF_INET6, family);
     struct sockaddr_in6 addr;
     memset(&addr, 0, sizeof(addr));
     addr.sin6_family = AF_INET6;
-    addr.sin6_addr = in6addr_any;
+    memset(&addr.sin6_addr, 0, sizeof(addr.sin6_addr));  // in6addr_any
     addr.sin6_port = htons(tcpport_);
     int tcprc = bind(tcpfd_, (struct sockaddr*)&addr, sizeof(addr));
     EXPECT_EQ(0, tcprc) << "Failed to bind AF_INET6 to TCP port " << tcpport_;
     addr.sin6_port = htons(udpport_);
     int udprc = bind(udpfd_, (struct sockaddr*)&addr, sizeof(addr));
     EXPECT_EQ(0, udprc) << "Failed to bind AF_INET6 to UDP port " << udpport_;
+    // retrieve system-assigned port
+    if (udpport_ == dynamic_port) {
+      ares_socklen_t len = sizeof(addr);
+      auto result = getsockname(udpfd_, (struct sockaddr*)&addr, &len);
+      EXPECT_EQ(0, result);
+      udpport_ = ntohs(addr.sin6_port);
+      EXPECT_NE(dynamic_port, udpport_);
+    }
+    if (tcpport_ == dynamic_port) {
+      ares_socklen_t len = sizeof(addr);
+      auto result = getsockname(tcpfd_, (struct sockaddr*)&addr, &len);
+      EXPECT_EQ(0, result);
+      tcpport_ = ntohs(addr.sin6_port);
+      EXPECT_NE(dynamic_port, tcpport_);
+    }
   }
   if (verbose) std::cerr << "Configured "
                          << (family == AF_INET ? "IPv4" : "IPv6")
@@ -207,6 +262,68 @@ MockServer::~MockServer() {
   sclose(udpfd_);
 }
 
+void MockServer::ProcessPacket(int fd, struct sockaddr_storage *addr, socklen_t addrlen,
+                               byte *data, int len) {
+
+  // Assume the packet is a well-formed DNS request and extract the request
+  // details.
+  if (len < NS_HFIXEDSZ) {
+    std::cerr << "Packet too short (" << len << ")" << std::endl;
+    return;
+  }
+  int qid = DNS_HEADER_QID(data);
+  if (DNS_HEADER_QR(data) != 0) {
+    std::cerr << "Not a request" << std::endl;
+    return;
+  }
+  if (DNS_HEADER_OPCODE(data) != O_QUERY) {
+    std::cerr << "Not a query (opcode " << DNS_HEADER_OPCODE(data)
+              << ")" << std::endl;
+    return;
+  }
+  if (DNS_HEADER_QDCOUNT(data) != 1) {
+    std::cerr << "Unexpected question count (" << DNS_HEADER_QDCOUNT(data)
+              << ")" << std::endl;
+    return;
+  }
+  byte* question = data + 12;
+  int qlen = len - 12;
+
+  char *name = nullptr;
+  long enclen;
+  ares_expand_name(question, data, len, &name, &enclen);
+  if (!name) {
+    std::cerr << "Failed to retrieve name" << std::endl;
+    return;
+  }
+  qlen -= enclen;
+  question += enclen;
+  std::string namestr(name);
+  ares_free_string(name);
+
+  if (qlen < 4) {
+    std::cerr << "Unexpected question size (" << qlen
+              << " bytes after name)" << std::endl;
+    return;
+  }
+  if (DNS_QUESTION_CLASS(question) != C_IN) {
+    std::cerr << "Unexpected question class (" << DNS_QUESTION_CLASS(question)
+              << ")" << std::endl;
+    return;
+  }
+  int rrtype = DNS_QUESTION_TYPE(question);
+
+  if (verbose) {
+    std::vector<byte> req(data, data + len);
+    std::cerr << "received " << (fd == udpfd_ ? "UDP" : "TCP") << " request " << PacketToString(req)
+              << " on port " << (fd == udpfd_ ? udpport_ : tcpport_) << std::endl;
+    std::cerr << "ProcessRequest(" << qid << ", '" << namestr
+              << "', " << RRTypeToString(rrtype) << ")" << std::endl;
+  }
+  ProcessRequest(fd, addr, addrlen, qid, namestr, rrtype);
+
+}
+
 void MockServer::ProcessFD(int fd) {
   if (fd != tcpfd_ && fd != udpfd_ && connfds_.find(fd) == connfds_.end()) {
     // Not one of our FDs.
@@ -229,6 +346,7 @@ void MockServer::ProcessFD(int fd) {
   int len = recvfrom(fd, BYTE_CAST buffer, sizeof(buffer), 0,
                      (struct sockaddr *)&addr, &addrlen);
   byte* data = buffer;
+
   if (fd != udpfd_) {
     if (len == 0) {
       connfds_.erase(std::find(connfds_.begin(), connfds_.end(), fd));
@@ -239,71 +357,26 @@ void MockServer::ProcessFD(int fd) {
       std::cerr << "Packet too short (" << len << ")" << std::endl;
       return;
     }
-    int tcplen = (data[0] << 8) + data[1];
-    data += 2;
-    len -= 2;
-    if (tcplen != len) {
-      std::cerr << "Warning: TCP length " << tcplen
-                << " doesn't match remaining data length " << len << std::endl;
+    /* TCP might aggregate the various requests into a single packet, so we
+     * need to split */
+    while (len) {
+      int tcplen = (data[0] << 8) + data[1];
+      data += 2;
+      len -= 2;
+      if (tcplen > len) {
+        std::cerr << "Warning: TCP length " << tcplen
+                  << " doesn't match remaining data length " << len << std::endl;
+      }
+      int process_len = (tcplen > len)?len:tcplen;
+      ProcessPacket(fd, &addr, addrlen, data, process_len);
+      len -= process_len;
+      data += process_len;
     }
+  } else {
+    /* UDP is always a single packet */
+    ProcessPacket(fd, &addr, addrlen, data, len);
   }
 
-  // Assume the packet is a well-formed DNS request and extract the request
-  // details.
-  if (len < NS_HFIXEDSZ) {
-    std::cerr << "Packet too short (" << len << ")" << std::endl;
-    return;
-  }
-  int qid = DNS_HEADER_QID(data);
-  if (DNS_HEADER_QR(data) != 0) {
-    std::cerr << "Not a request" << std::endl;
-    return;
-  }
-  if (DNS_HEADER_OPCODE(data) != ns_o_query) {
-    std::cerr << "Not a query (opcode " << DNS_HEADER_OPCODE(data)
-              << ")" << std::endl;
-    return;
-  }
-  if (DNS_HEADER_QDCOUNT(data) != 1) {
-    std::cerr << "Unexpected question count (" << DNS_HEADER_QDCOUNT(data)
-              << ")" << std::endl;
-    return;
-  }
-  byte* question = data + 12;
-  int qlen = len - 12;
-
-  char *name = nullptr;
-  long enclen;
-  ares_expand_name(question, data, len, &name, &enclen);
-  if (!name) {
-    std::cerr << "Failed to retrieve name" << std::endl;
-    return;
-  }
-  qlen -= enclen;
-  question += enclen;
-  std::string namestr(name);
-  free(name);
-
-  if (qlen < 4) {
-    std::cerr << "Unexpected question size (" << qlen
-              << " bytes after name)" << std::endl;
-    return;
-  }
-  if (DNS_QUESTION_CLASS(question) != ns_c_in) {
-    std::cerr << "Unexpected question class (" << DNS_QUESTION_CLASS(question)
-              << ")" << std::endl;
-    return;
-  }
-  int rrtype = DNS_QUESTION_TYPE(question);
-
-  if (verbose) {
-    std::vector<byte> req(data, data + len);
-    std::cerr << "received " << (fd == udpfd_ ? "UDP" : "TCP") << " request " << PacketToString(req)
-              << " on port " << (fd == udpfd_ ? udpport_ : tcpport_) << std::endl;
-    std::cerr << "ProcessRequest(" << qid << ", '" << namestr
-              << "', " << RRTypeToString(rrtype) << ")" << std::endl;
-  }
-  ProcessRequest(fd, &addr, addrlen, qid, namestr, rrtype);
 }
 
 std::set<int> MockServer::fds() const {
@@ -359,7 +432,8 @@ MockChannelOptsTest::NiceMockServers MockChannelOptsTest::BuildServers(int count
   NiceMockServers servers;
   assert(count > 0);
   for (int ii = 0; ii < count; ii++) {
-    std::unique_ptr<NiceMockServer> server(new NiceMockServer(family, base_port + ii));
+    int port = base_port == dynamic_port ? dynamic_port : base_port + ii;
+    std::unique_ptr<NiceMockServer> server(new NiceMockServer(family, port));
     servers.push_back(std::move(server));
   }
   return servers;
@@ -381,9 +455,9 @@ MockChannelOptsTest::MockChannelOptsTest(int count,
   }
 
   // Point the library at the first mock server by default (overridden below).
-  opts.udp_port = mock_port;
+  opts.udp_port = server_.udpport();
   optmask |= ARES_OPT_UDP_PORT;
-  opts.tcp_port = mock_port;
+  opts.tcp_port = server_.tcpport();
   optmask |= ARES_OPT_TCP_PORT;
 
   // If not already overridden, set short-ish timeouts.
@@ -482,7 +556,12 @@ void MockChannelOptsTest::Process() {
 std::ostream& operator<<(std::ostream& os, const HostResult& result) {
   os << '{';
   if (result.done_) {
-    os << StatusToString(result.status_) << " " << result.host_;
+    os << StatusToString(result.status_);
+    if (result.host_.addrtype_ != -1) {
+      os << " " << result.host_;
+    } else {
+      os << ", (no hostent)";
+    }
   } else {
     os << "(incomplete)";
   }
@@ -493,8 +572,10 @@ std::ostream& operator<<(std::ostream& os, const HostResult& result) {
 HostEnt::HostEnt(const struct hostent *hostent) : addrtype_(-1) {
   if (!hostent)
     return;
+
   if (hostent->h_name)
     name_ = hostent->h_name;
+
   if (hostent->h_aliases) {
     char** palias = hostent->h_aliases;
     while (*palias != nullptr) {
@@ -502,7 +583,9 @@ HostEnt::HostEnt(const struct hostent *hostent) : addrtype_(-1) {
       palias++;
     }
   }
+
   addrtype_ = hostent->h_addrtype;
+
   if (hostent->h_addr_list) {
     char** paddr = hostent->h_addr_list;
     while (*paddr != nullptr) {
@@ -514,9 +597,11 @@ HostEnt::HostEnt(const struct hostent *hostent) : addrtype_(-1) {
 }
 
 std::ostream& operator<<(std::ostream& os, const HostEnt& host) {
-  os << '{';
-  os << "'" << host.name_ << "' "
-     << "aliases=[";
+  os << "{'";
+  if (host.name_.length() > 0) {
+    os << host.name_;
+  }
+  os << "' aliases=[";
   for (size_t ii = 0; ii < host.aliases_.size(); ii++) {
     if (ii > 0) os << ", ";
     os << host.aliases_[ii];
@@ -535,14 +620,91 @@ std::ostream& operator<<(std::ostream& os, const HostEnt& host) {
 void HostCallback(void *data, int status, int timeouts,
                   struct hostent *hostent) {
   EXPECT_NE(nullptr, data);
+  if (data == nullptr)
+    return;
+
   HostResult* result = reinterpret_cast<HostResult*>(data);
   result->done_ = true;
   result->status_ = status;
   result->timeouts_ = timeouts;
-  result->host_ = HostEnt(hostent);
+  if (hostent)
+    result->host_ = HostEnt(hostent);
   if (verbose) std::cerr << "HostCallback(" << *result << ")" << std::endl;
 }
 
+std::ostream& operator<<(std::ostream& os, const AddrInfoResult& result) {
+  os << '{';
+  if (result.done_ && result.ai_) {
+    os << StatusToString(result.status_) << " " << result.ai_;
+  } else {
+    os << "(incomplete)";
+  }
+  os << '}';
+  return os;
+}
+
+std::ostream& operator<<(std::ostream& os, const AddrInfo& ai) {
+  os << '{';
+  if (ai == nullptr) {
+    os << "nullptr}";
+    return os;
+  }
+
+  struct ares_addrinfo_cname *next_cname = ai->cnames;
+  while(next_cname) {
+    if(next_cname->alias) {
+      os << next_cname->alias << "->";
+    }
+    if(next_cname->name) {
+      os << next_cname->name;
+    }
+    if((next_cname = next_cname->next))
+      os << ", ";
+    else
+      os << " ";
+  }
+
+  struct ares_addrinfo_node *next = ai->nodes;
+  while(next) {
+    //if(next->ai_canonname) {
+      //os << "'" << next->ai_canonname << "' ";
+    //}
+    unsigned short port = 0;
+    os << "addr=[";
+    if(next->ai_family == AF_INET) {
+      sockaddr_in* sin = (sockaddr_in*)next->ai_addr;
+      port = ntohs(sin->sin_port);
+      os << AddressToString(&sin->sin_addr, 4);
+    }
+    else if (next->ai_family == AF_INET6) {
+      sockaddr_in6* sin = (sockaddr_in6*)next->ai_addr;
+      port = ntohs(sin->sin6_port);
+      os << "[" << AddressToString(&sin->sin6_addr, 16) << "]";
+    }
+    else
+      os << "unknown family";
+    if(port) {
+      os << ":" << port;
+    }
+    os << "]";
+    if((next = next->ai_next))
+      os << ", ";
+  }
+  os << '}';
+  return os;
+}
+
+void AddrInfoCallback(void *data, int status, int timeouts,
+                      struct ares_addrinfo *ai) {
+  EXPECT_NE(nullptr, data);
+  AddrInfoResult* result = reinterpret_cast<AddrInfoResult*>(data);
+  result->done_ = true;
+  result->status_ = status;
+  result->timeouts_= timeouts;
+  result->ai_ = AddrInfo(ai);
+  if (verbose) std::cerr << "AddrInfoCallback(" << *result << ")" << std::endl;
+}
+
 std::ostream& operator<<(std::ostream& os, const SearchResult& result) {
   os << '{';
   if (result.done_) {
@@ -663,5 +825,15 @@ TempFile::TempFile(const std::string& contents)
 
 }
 
+VirtualizeIO::VirtualizeIO(ares_channel c)
+  : channel_(c)
+{
+  ares_set_socket_functions(channel_, &default_functions, 0);
+}
+
+VirtualizeIO::~VirtualizeIO() {
+  ares_set_socket_functions(channel_, 0, 0);
+}
+
 }  // namespace test
 }  // namespace ares
